-- A set of different-looking buff auras. Inspired by Elkano's Buff Bars.
-- The player can left-click on a buff to cancel it.

-- Set our default placement values. If a SavedVariable file exists, then these will be overwritten after all our code is loaded. If it doesn't, then these will work as defaults.


-- Create our context so we can do rendering.
AC = {}
AC.vers = "1.0.5"
AC.context = UI.CreateContext("ACContext")

AC.frames = {}

AC.framesLocked = true
AC.debugging = false
AC.initialised = false

AC.config = ACConfig

FrameTypes = {
	"buffFrame", "debuffFrame"
}
local function format(...)
	return string.format(...)
end

local function p(...)
	if( AC.debugging == true) then
		print(...)
	end
end

local function e(data)
	if type(data) == "table" then
		for i,v in ipairs(data) do
			t = "+"..e(v)
			return t
		end
	else
		t = "-"..data
		return (t)
	end
end

AC.lastTick = nil

local function throttle()
	local cTime = time
	p(cTime)
end

local function timeShort(remaining)
	if remaining >= 3600 then
		return (string.format("%dh", math.ceil(remaining / 3600)))
	elseif remaining >= 60 then
		return (string.format("%dm", math.floor(remaining / 60) % 60))
	else
		return (string.format("%d", math.floor(remaining) % 60))
	end
end
 
 local function setDefaults(unitName, frameType)
	AC.coords[unitName] = AC.coords[unitName] or {}
	AC.coords[unitName][frameType] = AC.coords[unitName][frameType] or { x = 300, y = 300 }
 end
 
local function createFrames()
	for  unitName, _ in pairs(AC.config) do
		AC.frames[unitName] = {}
		for frameType, _ in pairs(AC.config[unitName]) do
			--p(AC.config[unitName][frameType].corner[1])
			p("cf: "..unitName.." - "..frameType)
			AC.frames[unitName][frameType] = {}
			
			local corner = AC.config[unitName][frameType].corner[1]..AC.config[unitName][frameType].corner[2]
			
			local frame = UI.CreateFrame("Frame", frameType, AC.context)
			setDefaults(unitName, frameType)
			frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", AC.coords[unitName][frameType].x, AC.coords[unitName][frameType].y)
			
			local text = UI.CreateFrame("Text", "Name", frame)
			text:SetLayer(1)
			text:SetText(unitName.." - "..frameType)
			text:SetWidth(text:GetFullWidth())
			text:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT")
			text:SetFont("AuraControl", ACGlobals.font)
			text:SetFontSize(24)
			text:SetVisible(false)
			frame.text = text
			--frame:SetBackgroundColor(0,0,0,.5)
			function frame.Event:LeftDown()
				p("frame clicked", unitName, frameType)
				-- XXX can't actually "set" the current target from what I can gather
				--if(unitName == "player") then
				--	print("clicky")
				--	setUnitFrameTarget("player.target", "player")
				--	setUnitFrameTarget("player.target.target", "player")
				--end
						
				if(AC.framesLocked == false) then
					self.MouseDown = true
					mouseData = Inspect.Mouse()
					self.MyStartX = frame:GetLeft()
					self.MyStartY = frame:GetTop()
					self.StartX = mouseData.x - self.MyStartX
					self.StartY = mouseData.y - self.MyStartY
					tempX = frame:GetLeft()
					tempY = frame:GetTop()
					tempW = frame:GetWidth()
					tempH =	frame:GetHeight()
					frame:ClearAll()
					frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", tempX, tempY)
					frame:SetWidth(tempW)
					frame:SetHeight(tempH)
					self:SetBackgroundColor(0.3,0.0,0.0,0.5)
				end
			end
			function frame.Event:MouseMove()
				if self.MouseDown then
					local newX, newY
					mouseData = Inspect.Mouse()
					newX = mouseData.x - self.StartX
					newY = mouseData.y - self.StartY
					frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", newX, newY)
				end
			end
			function frame.Event:LeftUp()
				if self.MouseDown then
					self.MouseDown = false
					frame:SetBackgroundColor(0.0, 0.0, 0.0, 0.5)
								
					-- store frame placement in saved var
					AC.coords[unitName][frameType].x = frame:GetLeft()
					AC.coords[unitName][frameType].y = frame:GetTop()
				end
			end
			
			AC.frames[unitName][frameType]["frame"] = frame 
			AC.frames[unitName][frameType]["activeAuras"] = {}
			AC.frames[unitName][frameType]["zombieAuras"] = {}	
			AC.frames[unitName][frameType]["last"] = nil
			AC.frames[unitName][frameType]["first"] = nil
			AC.frames[unitName][frameType]["corner"] = nil
			AC.frames[unitName][frameType]["lineCount"] = 0
		end
	end
end

local function resetAuras(unitName)  -- Hide everything we have and start adding auras one at a time.
	for i, frameType in ipairs(FrameTypes) do
		p("resetting ".. frameType .." aura on ".. unitName)
		for _, aura in pairs(AC.frames[unitName][frameType]["activeAuras"]) do
			table.insert(AC.frames[unitName][frameType]["zombieAuras"], aura)
			aura:SetVisible(false)
			--aura:SetAllPoints()
			--aura:SetPoint(AC.frames[unitName][frameType].lCorner, AC.frames[unitName][aura.frameType]["frame"], AC.frames[unitName][frameType].lCorner)
		end
		AC.frames[unitName][frameType]["activeAuras"] = {}
		AC.frames[unitName][frameType]["last"] = nil
		AC.frames[unitName][frameType]["first"] = nil
		AC.frames[unitName][frameType]["corner"] = nil
		AC.frames[unitName][frameType]["lineCount"] = 0
	end
end

local function addAura(unitName, buff, time)
	local frameType
	if buff.debuff then
		frameType = "debuffFrame"
	else
		frameType = "buffFrame"
	end
	p("ca: "..unitName.." - "..frameType.." : "..buff.name)
	-- attempt to reuse an old aura
	local aura = table.remove(AC.frames[unitName][frameType]["zombieAuras"])
	--local AC.config = AC.config[unitName] and AC.config[unitName][frameType] or nil
	local corner = AC.config[unitName][frameType].corner[1]..AC.config[unitName][frameType].corner[2]
	if aura then

		--aura:SetHeight(aura.count:GetFullHeight())
		-- Set location
		p(unitName, frameType)
		aura.frameType = frameType
		aura.frameFrom = AC.config[unitName][frameType].frameFrom
		aura.frameTo = AC.config[unitName][frameType].frameTo
		aura.iconFrom = AC.config[unitName][frameType].iconFrom
		aura.iconTo = AC.config[unitName][frameType].iconTo
		aura:SetPoint(corner, AC.frames[unitName][aura.frameType]["frame"], corner)
	else
		-- We don't have any auras remaining, so we create a new one.
		-- Our Bars are considered single objects that can be dealt with atomically. Each one has the functionality needed to update itself.
		
		--aura:SetHeight(aura.count:GetFullHeight())
		-- Set location
		aura = UI.CreateFrame("Frame", "Bar", AC.frames[unitName][frameType]["frame"])
		aura.frameType = frameType
		aura.frameFrom = AC.config[unitName][frameType].frameFrom
		aura.frameTo = AC.config[unitName][frameType].frameTo
		aura.iconFrom = AC.config[unitName][frameType].iconFrom
		aura.iconTo = AC.config[unitName][frameType].iconTo
		aura:SetPoint(corner, AC.frames[unitName][aura.frameType]["frame"], corner)
		
		aura.count = UI.CreateFrame("Text", "Count", aura)
		aura.countS = UI.CreateFrame("Text", "CountS", aura)
		
		aura.timer = UI.CreateFrame("Text", "Timer", aura)
		aura.timerS = UI.CreateFrame("Text", "TimerS", aura)
		aura.icon = UI.CreateFrame("Texture", "Icon", aura)

		-- Solid background - this is the actual "aura" part of it.
		

		aura.icon:SetPoint("TOPLEFT", aura, "TOPLEFT") -- The icon is pinned to the top-left corner of the aura.
		aura.icon:SetPoint("BOTTOM", aura, "BOTTOM") -- Vertically, it always fills the entire aura.

		aura.count:SetFont("AuraControl", ACGlobals.font)
		aura.countS:SetFont("AuraControl", ACGlobals.font)
		aura.count:SetText("x10")
		aura.countS:SetText("x10")
		aura.count:SetFontSize(ACGlobals.fontSize*AC.config[unitName][frameType].iconScale)
		aura.countS:SetFontSize(ACGlobals.fontSize*AC.config[unitName][frameType].iconScale)
		aura.count:SetHeight(aura.count:GetFullHeight())
		aura.countS:SetHeight(aura.count:GetFullHeight())
		aura.count:SetWidth(aura.count:GetFullWidth())
		aura.countS:SetWidth(aura.count:GetFullWidth())
		aura.count:SetPoint("TOPRIGHT", aura.icon, "TOPRIGHT", 0, 0)
		aura.countS:SetPoint("TOPRIGHT", aura.icon, "TOPRIGHT", 1, 2)
		aura.count:SetLayer(2)
		aura.countS:SetLayer(1)
		aura.countS:SetFontColor(0,0,0,1)
		
		


		aura.timer:SetFont("AuraControl", ACGlobals.font)
		aura.timerS:SetFont("AuraControl", ACGlobals.font)
		aura.timer:SetText("19m")
		aura.timerS:SetText("19m")
		aura.timer:SetFontSize(ACGlobals.fontSize*AC.config[unitName][frameType].iconScale)
		aura.timerS:SetFontSize(ACGlobals.fontSize*AC.config[unitName][frameType].iconScale)
		aura.timer:SetHeight(aura.timer:GetFullHeight())
		aura.timerS:SetHeight(aura.timer:GetFullHeight())
		aura.timer:SetPoint("BOTTOMRIGHT", aura.icon, "BOTTOMRIGHT", 0, 4) -- The timer is pinned to the top-right corner.
		aura.timerS:SetPoint("BOTTOMRIGHT", aura.icon, "BOTTOMRIGHT", 1, 6)
		aura.timer:SetLayer(2)
		aura.timerS:SetLayer(1)
		aura.timerS:SetFontColor(0,0,0,1)
		--aura.count:SetPoint("TOPLEFT", aura.timer, "TOPLEFT") -- Make sure the text doesn't overrun the timer. We'll be changing the text's height based on the contents, but we'll leave the width calculated this way.

		aura:SetBackgroundColor(0,0,0,.75)
		aura:SetWidth(40*AC.config[unitName][frameType].iconScale)--ACConfig.unitFrameBarWidth + ACConfig.unitFrameOffset*2)
		
		--
		-- Set Buff - requires a buff and a timestamp
		--
		function aura:SetBuff(buff, time)
		self:SetHeight(40 * AC.config[unitName][frameType].iconScale)


		  -- Re-square and set the icon
		  self.icon:SetWidth(self.icon:GetHeight())
		  self.icon:SetTexture("Rift", buff.icon)

		  -- Display our stacking multiple.
		  if buff.stack then
			self.count:SetText("x" .. buff.stack)
			self.countS:SetText("x" .. buff.stack)
		  else
			self.count:SetText("")
			self.countS:SetText("")
		  end
		  
		  if buff.duration then
			self.completion = buff.begin + buff.duration
			self.duration = buff.duration
			
			-- Display everything we might have hidden.
			
			self.timer:SetVisible(true)
			self.timerS:SetVisible(true)
			
			self:Tick(time)
		  else
			self.completion = nil
			
			-- This is a permanent buff without a timer, so don't show any of that.
			
			self.timer:SetVisible(false)
			self.timer:SetWidth(0)
			self.timerS:SetVisible(false)
			self.timerS:SetWidth(0)
		  end
		  
		  self.debuff = buff.debuff
		  self.buffid = buff.id
		end
		
		function aura:UpdateBlink(self)
			if (self.Warning) then
				if (self:GetVisible()) then
					local a = self:GetAlpha();
					
					if (self.Blink) then
						a = a + 0.1;
					else
						a = a - 0.1;
					end
					
					self:SetAlpha(minmax(a, 0.25, 1)); 
					if (a >= 1) then
						self.Blink = false;
					elseif (a <= 0.29) then
						self.Blink = true;
					end
					
				end
			else
				self.Blink = false;
			end
		end
		
		--
		-- This is our update function, called once every frame.
		--
		function aura:Tick(time)
			if self.completion then
				local remaining = self.completion - time

				if remaining < 0 then
					self.timer:SetVisible(false)
					self.timerS:SetVisible(false)
				else
					self.timer:SetText(timeShort(remaining))
					self.timerS:SetText(timeShort(remaining))
					self.timer:SetWidth(self.timer:GetFullWidth())
					self.timerS:SetWidth(self.timer:GetFullWidth())
					if remaining <= 59*60 then
					--	self:SetHeight(48)
					--	self.icon:SetWidth(48)
					end
				end
			end
		end
		
		--
		-- Finally, if we're clicked, we want to cancel whatever buff is on us.
		--
		function aura.Event:RightDown()
		  Command.Buff.Cancel(self.buffid)
		end
	end
	
	table.insert(AC.frames[unitName][aura.frameType]["activeAuras"], aura)

	-- Show the aura and set the data.
	
	aura:SetBuff(buff, time)
	--print (unitName, aura.frameType, aura.frameFrom, aura.frameTo)
	--local corner = AC.config[unitName][frameType].corner[1]..AC.config[unitName][frameType].corner[2]
	local iconGrowth = ""
	local iFrom, iTo, iAdj, lAdj
	if AC.config[unitName][frameType].corner[1] == "TOP" then
		lCorner = "BOTTOM"..AC.config[unitName][frameType].corner[2]
		lAdj = 1
	elseif AC.config[unitName][frameType].corner[1] == "BOTTOM" then
		lCorner = "TOP"..AC.config[unitName][frameType].corner[2]
		lAdj = -1
	end
	if AC.config[unitName][frameType].corner[2] == "LEFT" then
		iFrom = AC.config[unitName][frameType].corner[1].."LEFT"
		iTo = AC.config[unitName][frameType].corner[1].."RIGHT"
		iAdj = 1
	elseif AC.config[unitName][frameType].corner[2] == "RIGHT" then
		iFrom = AC.config[unitName][frameType].corner[1].."RIGHT"
		iTo = AC.config[unitName][frameType].corner[1].."LEFT"
		iAdj = -1
	end
	p(unitName, frameType, AC.frames[unitName][aura.frameType]["lineCount"])
	AC.frames[unitName][aura.frameType].lCorner = lCorner
	if AC.frames[unitName][aura.frameType]["lineCount"] == AC.config[unitName][aura.frameType].perLine then
		AC.frames[unitName][aura.frameType]["lineCount"] = 0
		aura:SetPoint(corner, AC.frames[unitName][aura.frameType]["corner"], lCorner, lAdj*0, lAdj*3)
		AC.frames[unitName][aura.frameType]["corner"] = aura
	elseif not AC.frames[unitName][aura.frameType]["last"] then
		aura:SetPoint(corner, AC.frames[unitName][aura.frameType]["frame"], corner, 0, 0)
		AC.frames[unitName][aura.frameType]["corner"] = aura
		AC.frames[unitName][aura.frameType]["first"] = aura
	else
		aura:SetPoint(iFrom, AC.frames[unitName][aura.frameType]["last"], iTo, iAdj*3, iAdj*0)
	end
	AC.frames[unitName][aura.frameType]["last"] = aura
	AC.frames[unitName][frameType]["lineCount"] = AC.frames[unitName][frameType]["lineCount"] + 1
	if not locked and AC.frames[unitName][aura.frameType]["corner"] == aura then
		aura:SetVisible(false)
	else
		aura:SetVisible(true)
	end
end

local function AddRemoveAuras(unitName, time)
	-- inspect buffs for unitName
	p("aa. "..unitName)
	local bufflist = Inspect.Buff.List(unitName)

	-- If we don't get anything, then we don't currently have information about the player.
	-- This may happen when the player is logging in or teleporting long distances.
	if bufflist then  
	
		local buffdetails = Inspect.Buff.Detail(unitName, bufflist)
		resetAuras(unitName)
		
		-- We want to order buffs by their time remaining
		-- splitting apart buffs and debuffs.
		local bauras = {}
		for id, buff in pairs(buffdetails) do
		buff.id = id  -- Make a copy of the ID, because we'll need it
		table.insert(bauras, buff)
		end

		-- sort on time
		table.sort(
			bauras, function (a, b)
				if a.debuff ~= b.debuff then
				  return b.debuff
				end
				
				if a.duration and b.duration then return a.remaining > b.remaining end
				if not a.duration and not b.duration then return false end
				return not a.duration
			end
		)
		
		-- debug
		p("buff AC.configuration for ", unitName)
		p("show player cast buffs only?", ACGlobals.showPlayerBuffsOnly[unitName])
		p("show all buffs", ACGlobals.showAllBuffs[unitName])
		p("show player cast DEbuffs only?", ACGlobals.showPlayerDebuffsOnly[unitName])
		p("show all DEbuffs?", ACGlobals.showAllDebuffs[unitName])

		
		
		-- Now that we have the ordering, we just add the auras one at a time. Done!
		for k, buff in ipairs(bauras) do
			-- if the buff is a debuff...
			if buff.debuff then frameType = "debuffFrame" else frameType = "buffFrame" end
			p(unitName, frameType)
			if (buff.duration == nil and unitName == "player") or (AC.config[unitName] == nil) or (buff.duration and buff.duration > AC.config[unitName][frameType].maxTime) then
				
			elseif ( buff.debuff ) then
				-- if the frame is AC.configured to show only player cast debuffs
				if(ACGlobals.showPlayerDebuffsOnly[unitName] == true) then
					if (buff.caster == Inspect.Unit.Lookup("player")) then
						addAura(unitName, buff, time)
					end
				-- if the frame is AC.configured to show all debuffs
				elseif(ACGlobals.showAllDebuffs[unitName] == true) then
					addAura(unitName, buff, time)
				end
			-- the buff is a buff :P
			else
				-- if the frame is AC.configured to show only player cast buffs
				if(ACGlobals.showPlayerBuffsOnly[unitName] == true) then
					if (buff.caster == Inspect.Unit.Lookup("player")) then
						addAura(unitName, buff, time)
					end
				-- if the frame is AC.configured to show all buffs
				elseif(ACGlobals.showAllBuffs[unitName] == true) then
					addAura(unitName, buff, time)
				end
			end
		end
	else
		resetAuras(unitName) 
	end
end

local function refresh(time)
	-- check all the frames we have / add buffs as required
	p("Refresh")
	for unitName, _ in pairs(AC.frames) do
		AddRemoveAuras(unitName, time)
	end
end

local function setFrameLocation(unitName, frameType, x, y)
	AC.frames[unitName][frameType]["frame"]:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x,y )
end

local function loadSavedFrameLocations()
	for unitName, _ in pairs(AC.coords) do
		for frameType, _ in pairs(AC.coords[unitName]) do
			if AC.frames[unitName] then
				setFrameLocation(unitName, frameType, AC.coords[unitName][frameType].x, AC.coords[unitName][frameType].y)
			end
		end
	end
	refresh(Inspect.Time.Frame())
end

local function createDummies()
	for  unitName, _ in pairs(AC.config) do
		for frameType, _ in pairs(AC.config[unitName]) do
			local perLine = AC.config[unitName][frameType].perLine
			p("cd. "..unitName.." - "..frameType.." : "..perLine)
			local aura = {}
			aura.name = "dummy"
			if frameType == "buffFrame" then
				aura.debuff = false
			else
				aura.debuff = true
			end
			for i=1,perLine do
				addAura(unitName, aura, time)
			end
		end
	end
end

-- setup frames
local function initialiseFrames()
	p("initialising frames")

	createFrames()
	--createDummies()
	-- load saved settings
	loadSavedFrameLocations()
	
	-- updates the unit frame details
	--updateFrames()
end

local function targetChanged()
	p("target changed")
	--updateUnitFrames()
	AC.resyncBuffs = true
end

-- lock frames
locked = true
local function lockFrames()
	print("Frames Locked")
	print("/ac unlock : unlock frames for placement")
	AC.framesLocked = true
	for unitName, _ in pairs(AC.frames) do
		for frameType, _ in pairs(AC.frames[unitName]) do
			AC.frames[unitName][frameType]["frame"].text:SetVisible(false)
			AC.frames[unitName][frameType]["frame"]:SetBackgroundColor(0,0,0,0)
			if AC.frames[unitName][frameType]["first"] then
				AC.frames[unitName][frameType]["first"]:SetVisible(true)
				locked = true
			end
		end
	end
end
-- unlock frames
local function unlockFrames()
	print("Frames UnLocked")
	print("/ac lock   : lock frames in place")
	AC.framesLocked = false
	for unitName, _ in pairs(AC.frames) do
		for frameType, _ in pairs(AC.frames[unitName]) do
			AC.frames[unitName][frameType]["frame"].text:SetVisible(true)
			AC.frames[unitName][frameType]["frame"]:SetBackgroundColor(0,0,0,1)
			if AC.frames[unitName][frameType]["first"] then
				AC.frames[unitName][frameType]["first"]:SetVisible(false)
				locked = false
			end
		end
	end
end

local function resetFrames()
	print("Frames Reset")
	print("/ac lock   : lock frames in place")
	
	-- reset positioning
	AC.config = ACConfig
	AC.coords = ACFramePlacement
	
	loadSavedFrameLocations()
	unlockFrames()
	
end

local resynch = false
local function tick()
	if (AC.initialised == false) then
		p("do init")
		initialiseFrames()
		AC.resyncBuffs = true
		AC.initialised = true
	end
	if AC.resyncBuffs then
	-- A recalculation has been queued, so go ahead and recalculate.
		refresh(Inspect.Time.Frame())
		AC.resyncBuffs = false
	else
		-- Just do a tick refresh
		local time = Inspect.Time.Frame()
		for i, frame in ipairs(FrameTypes) do
			for unitName, value in pairs(AC.frames) do
				for _, v in ipairs(AC.frames[unitName][frame]["activeAuras"]) do
					v:Tick(time)
				end
			end
		end
	end
end

local function saveConfig(addon)
	--p("saveConfig")
	if addon == "AuraControl" then
		FrameConfig = AC.config
		FramePlacement = AC.coords
	end
end
local firstrun = false
local function loadConfig(addon)
	--p("loadConfig")
	if addon == "AuraControl" then
		if FramePlacement then
			AC.coords = FramePlacement
		else
			print("First Run --- setting defaults.")
			AC.coords = ACFramePlacement
			firstRun = true
		end
	end
end

local function addonLoaded(addon)
	if(addon == "AuraControl") then
		print("vers "..AC.vers)
		if AC.config then
			p("C.")
		end
	end
end

local function updateAuraFrames()
	for key, value in pairs(AC.frames) do
		p("updating ",key)
		--updateUnitFrame(key)
	end
end

local function targetChanged()
	p("target changed")
	AC.resyncBuffs = true
end

local function commandLine(args)
	if args == "lock" then
		lockFrames()
	elseif args == "unlock" then
		unlockFrames()
	elseif args == "reset" then
		resetFrames()
	else
		print("vers. 1.0")
		print("commandline error  ---")
		print("/ac unlock : unlock frames for placement")
		print("/ac lock   : lock frames in place")
	end
end

local function startup()
	table.insert(Event.Buff.Add, {function () AC.resyncBuffs = true end, "AuraControl", "refresh"})
	table.insert(Event.Buff.Change, {function () AC.resyncBuffs = true end, "AuraControl", "refresh"})
	table.insert(Event.Buff.Remove, {function () AC.resyncBuffs = true end, "AuraControl", "refresh"})
	
	table.insert(Event.Ability.Target, {targetChanged, "AuraControl", "updateAuraFrames"})
	
	table.insert(Event.Addon.Load.End, {addonLoaded, "AuraControl", "addonLoaded"})
	
	-- Our update event. This triggers every single frame.
	table.insert(Event.System.Update.Begin, {tick, "AuraControl", "refresh"})
	
	table.insert(Event.Addon.SavedVariables.Save.Begin, {saveConfig, "AuraControl", "Save variables"})
	table.insert(Event.Addon.SavedVariables.Load.End, {loadConfig, "AuraControl", "Load variables"})

	table.insert(Command.Slash.Register("ac"), {commandLine, "AuraControl", "AuraControl.command"})
	table.insert(Command.Slash.Register("auracontrol"), {commandLine, "AuraControl", "AuraControl.command"})

end

startup()